#-----------------------------------------------------------------------------
# Eureka 3 Modernized!
# Aug. 29, 2004
# Try for Esther
# Copyright (c) 2004, David G. Drumright (ockham)

#-----------------------------------------------------------------------------

from Tkinter import *
root = Tk()

import math
import string
import poser
scene = poser.Scene()


# Global vars
global SoftName
global SoftObject
global SoftGeom
global SoftPolyCount
global HardNames
global HardObjects
global SoftLevel
global SoftVertCount
global SmoothIndexes
global SmoothYvals
global SmoothOn


SmoothIndexes	= []
SmoothYvals 	= []
HardNames = []	# List of tuples (FigName,ActorName)
SoftName = ('','')   # Just one tuple (FigName, ActorName)
HardObjects=[]

#-----------------------------------------------------------------------------
# Method for figuring whether a point is inside a 2D polygon.  Found
# on web, credited to Ben Wyatt.  Had to change the sign from
# original; Poser's winding order is apparently opposite to the usual.
#-----------------------------------------------------------------------------
#
def TriArea(x1, z1, x2, z2, x3, z3):
	# the function to calculate double the area of a triangle
	return (x2-x1)*(z3-z1) - (x3-x1)*(z2-z1)

#--------------------------------------------------------------

def PtInPoly(TryX, TryZ, Px, Pz, Count):
	# Px and Pz are arrays holding the X and Z points.
	for i in range(Count):	# for every triangle formed
		j = (i+1) % Count
		area = TriArea(TryX, TryZ, Px[j], Pz[j], Px[i], Pz[i]);
		# if area is negative, we know point is outside.
		if (area > 0.0):
		   return 0

	# Finished loop without breaking: Pt is in rect.
	return 1

#-----------------------------------------------------------------------------

def SetupVars():
	global SoftName
	global SoftObject
	global SoftGeom
	global HardNames
	global HardObjects
	global SoftVertCount
	global HardVertCount
	global SoftLevel
	global SmoothIndexes
	global SmoothYvals

	# See if the SoftObject is part of a figure.
	(FigName,ActorName) = SoftName

	if string.count(FigName,'[prop]'):
		SoftObject=scene.Actor(ActorName)
	else:
		TryFig=scene.Figure(FigName)
		SoftObject=TryFig.Actor(ActorName)
	try:
		SoftGeom = SoftObject.Geometry()
	except:
		raise 'No Soft geometry'
	try:
		SoftVertCount = SoftGeom.NumVertices()
	except:
		raise 'No Soft vert count'

	# Find the top Soft level so we can get a broad sense
	# of which part of the Hard is pushing in.
	SoftLevel = -10000000.0
	for i in range(SoftVertCount):
		OneWV = SoftGeom.WorldVertex(i)
		if OneWV.Y() > SoftLevel:
		   SoftLevel = OneWV.Y()


	# Turn the HardNames list into HardObjects.
	for (OneFigName,OneActorName) in HardNames:
		if string.count(OneFigName,'[prop]'):
			OneObj=scene.Actor(OneActorName)
		else:
			OneFig=scene.Figure(OneFigName)
			OneObj=OneFig.Actor(OneActorName)
		try:  #If this obj has Geom, add it to list.  If not, just skip
			TestGeom=OneObj.Geometry()
			HardObjects.append(OneObj)
		except:
			continue

#-----------------------------------------------------------------------------

def Displace():
	# This is the main action
	global SoftObject
	global SoftGeom
	global HardObject
	global SoftVertCount
	global SoftLevel
	global SmoothIndexes
	global SmoothYvals
	Changed = 0  # flag for running Preserve

	# Try a total-scaling approach for the Vertex.Y to WorldVertex.Y conversion.
	# The net Y scale should be: figure scale * figure Yscale * part scale * part Yscale.
	SoftFigScale = 1.0
	SoftFigYscale = 1.0
	if SoftObject.ItsFigure():
	   SoftBODY = SoftObject.ItsFigure().Actor('BODY')
	   SoftFigScale = SoftBODY.ParameterByCode(poser.kParmCodeASCALE).Value()
	   SoftFigYscale = SoftBODY.ParameterByCode(poser.kParmCodeYSCALE).Value()
	SoftScale  = SoftObject.ParameterByCode(poser.kParmCodeASCALE).Value()
	SoftYscale = SoftObject.ParameterByCode(poser.kParmCodeYSCALE).Value()
	SoftNetScale = SoftFigScale * SoftFigYscale * SoftScale * SoftYscale

	for OneHardObject in HardObjects:  # Biggest loop of all
		HardGeom = OneHardObject.Geometry()
		if not HardGeom: continue  # for theBODY or other dummy parts
		OffsetArray  = OneHardObject.WorldDisplacement()
		CenterX = OffsetArray[0]
		CenterY = OffsetArray[1]
		CenterZ = OffsetArray[2]

		MinHardX = 1000000.0
		MaxHardX = -1000000.0
		MinHardY = 1000000.0
		MaxHardY = -1000000.0
		MinHardZ = 1000000.0
		MaxHardZ = -1000000.0
		HardVertCount = HardGeom.NumVertices()
		for i in range(HardVertCount):
			BWV = HardGeom.WorldVertex(i)
			bwX = BWV.X()
			bwY = BWV.Y()
			bwZ = BWV.Z()
			# Set min and max of all three dims for the Hard
			if bwX < MinHardX:
			   MinHardX = bwX
			if bwX > MaxHardX:
			   MaxHardX = bwX
			if bwY < MinHardY:
			   MinHardY = bwY
			if bwY > MaxHardY:
			   MaxHardY = bwY
			if bwZ < MinHardZ:
			   MinHardZ = bwZ
			if bwZ > MaxHardZ:
			   MaxHardZ = bwZ
		# We now have the horizontal bounding 'cube' of the Hard.

		# Now run main loop across all Soft verts.
		# Ignore those outside the Hard's bounding rect.
		# Once we know we're close to the Hard:
		# Run the Hard's facets.

		# Clear Smooth before appending.
		Smooth = []

		for SoftIndex in range(SoftVertCount):
			OneWV = SoftGeom.WorldVertex(SoftIndex)
			SoftX = OneWV.X()
			SoftY = OneWV.Y()
			SoftZ = OneWV.Z()
			# Immediately exclude Soft outside the bounding rect.
			if	 SoftZ > MaxHardZ:
			   continue
			elif SoftZ < MinHardZ:
			   continue
			elif SoftX > MaxHardX:
			   continue
			elif SoftX < MinHardX:
			   continue

			# If the vertex is outside the Hard's entire outline, ignore.
			# Reminder: this rect is not aligned with the Hard but with
			# true directions, so we can do these max/mins properly.
			# The rect will be larger, and thus slower, if the Hard is
			# pointing obliquely, but still somewhat faster than doing the poly
			# loop for all the Soft.
			#
			# We now have a Soft vert near the Hard.
			# For each Polygon (facet) of the Hard, locate its WorldVertices,
			# and use those to isolate and displace the Verts of the Soft
			# that should be displaced.
			for HardPoly in HardGeom.Polygons():
				PolyX = []	# Start a list of Xs for the PtInPoly function
				PolyZ = []	# Start a list of Zs for the PtInPoly function
				SetList = HardGeom.Sets()
				First = HardPoly.Start()
				SetCount = HardPoly.NumVertices()
				Last = HardPoly.Start() + SetCount
				MinPolyY = 1000000.0
				MaxPolyY = -1000000.0
				for HardIndex in SetList[First:Last]:
					# OneSet is a list of the vert indexes for this facet.
					# Use these to break into the WorldVerts:
					BWVinPoly = HardGeom.WorldVertex(HardIndex)
					#BWVinPoly = HardGeom.Vertex(HardIndex)
					RX = BWVinPoly.X()
					PolyX.append(RX)

					RZ = BWVinPoly.Z()
					PolyZ.append(RZ)

					RY = BWVinPoly.Y()
					if RY < MinPolyY:
					   MinPolyY = RY
					if RY > MaxPolyY:
					   MaxPolyY = RY

				# Now we have Min/MaxPolyY and the lists of X and Z.
				# Here we make the main decision about what to do with this facet.
				if PtInPoly(SoftX, SoftZ, PolyX, PolyZ, SetCount):
				   # Hold in Smooth, but don't repeat. After a vert has
				   # been placed in Smooth, don't hit it again, because
				   # the second hit will be looking at Soft that's already displaced.
				   if SoftIndex not in SmoothIndexes:
					  if (MinPolyY < SoftY): # Work from bottom surface of hard downward into soft
						 MoveDelta = SoftY - MinPolyY
						 # The idea is that Delta is how far the WorldVert needs to move.
						 # Set the nonworldVert by this same dif.
						 SV = SoftGeom.Vertex(SoftIndex)
						 OrigSVY = SV.Y()
						 # Try a total-scaling approach.
						 SV.SetY((OrigSVY - MoveDelta)*SoftNetScale)
						 SmoothIndexes.append(SoftIndex)
						 SmoothYvals.append(OrigSVY)
						 Changed = 1
				# End actions for polys that arlatt or below Soft level.
			# End loop across Hard polygons
		# End loop across Soft World verts
	return Changed # so we can tell whether to Preserve
# End method Displace

#-----------------------------------------------------------------------------

#Smoothing, with the following method:
# Run the polys of the soft thing.	Each time a poly includes
# a vertex that was placed in Smooth, compare the current
# value with the orig in Smooth.  Move the other verts of
# this poly by 1/2 that value.

def Smooth(smoothcount):
  global SoftGeom
  global SmoothIndexes
  global SmoothYvals
  # Temporary arrays so we don't put things into the
  # same place we're looking.
  TempSmoothIndexes = []
  TempSmoothYvals = []

  if smoothcount == 0:
	 return
  elif smoothcount < 0: # Turn the chamfer backwards
	 Half = -2.0
  else:
	 Half =  2.0

  realcount = int(abs(smoothcount))
  for i in range(realcount):
	  for SoftPoly in SoftGeom.Polygons():
		  SetList = SoftGeom.Sets()
		  First = SoftPoly.Start()
		  SetCount = SoftPoly.NumVertices()
		  Last = SoftPoly.Start() + SetCount
		  for SoftIndex in SetList[First:Last]:
			  if SoftIndex in SmoothIndexes:  # If this vertex-number has been changed:
				 ChangedY = SoftGeom.Vertex(SoftIndex).Y() # The changed value of this vertex.Y
				 # Reminder: we don't need the orig value of the changed vertex.
				 # Now look at the other verts in this polygon.  For each
				 # one that we haven't moved yet, move it halfway to the changed Y.
				 for LookIndex in SetList[First:Last]:
					 if LookIndex != SoftIndex:
						OrigY = SoftGeom.Vertex(LookIndex).Y()
						Delta = (ChangedY + OrigY) / Half
						SoftGeom.Vertex(LookIndex).SetY(Delta)
						# Add this to the changed verts so the morph will
						# be right, and also to 'recurse' the smoothing for
						# more than one stage.
						TempSmoothIndexes.append(LookIndex)
						TempSmoothYvals.append(Delta)

  # Now that we're done looking, add the temps to the mains.
  SmoothIndexes.extend(TempSmoothIndexes)
  SmoothYvals.extend(TempSmoothYvals)


#-----------------------------------------------------------------------------
# Spawn a morph for each change.
#-----------------------------------------------------------------------------

def Preserve():
	global SoftObject
	global SoftGeom
	SoftObject.MarkGeomChanged() # preserve the change
	# Form title, Eureka plus a number.  Check to see if
	# we've already placed some, and number sequentially.
	MaxNum = -10000000
	MorphTitle = ''
	for OneParm in SoftObject.Parameters():
		OneName = OneParm.Name()
		if string.count(OneName,'Eureka'):
		   StringParts = string.split(OneName,'_')
		   ExistNum = int(StringParts[1])
		   if ExistNum > MaxNum:
			  ExistNum = MaxNum
	if MaxNum > 0: # found one
	   MorphTitle = 'Eureka_' + str(MaxNum + 1)
	else: # this is first, use 1
	   MorphTitle = 'Eureka_' + str(1)

	#Add that morph to the Soft thing
	SoftObject.SpawnTargetFromGeometry(SoftGeom,MorphTitle)
	# Fetch the just-made parm, then fill it with deltas where
	# we have them.  The Smooth array gives us this.
	NewParm = SoftObject.Parameter(MorphTitle)
	NewParm.SetValue(1.0) # Value init's to 0, should be 1.
	i = 0
	for VertIndex in SmoothIndexes:
		# The actual vertex is changed.  The delta is the actual
		# minus the preserved orig in Smooth.
		DeltaY = SoftGeom.Vertex(VertIndex).Y() - SmoothYvals[i]
		# Set deltas. Our X and Z DELTAS are always zero.
		NewParm.SetMorphTargetDelta(VertIndex, 0.0,DeltaY,0.0)
		i = i + 1

#------------------------------------------------------------
# Next section is the TK interface defs
#------------------------------------------------------------

class App:
	def __init__(self, master, textMessage):
		self.SmoothVar = IntVar()
		self.UpwardVar = IntVar()

		self.master = master
		master.title("Eureka")
		self.StatusEntry = Entry(self.master,width=50)
		self.StatusEntry.grid(row=0,column=0,columnspan=2)
		self.StatusEntry.insert(0,"Pick hard and soft, set values, hit Go.")

		self.SoftListFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
		self.SoftListFrame.grid(  row=1,column=0,pady=1)
		self.HardListFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
		self.HardListFrame.grid(  row=1,column=1,pady=1)
		self.EntryFrame = Frame(self.master,borderwidth=2,relief=RIDGE)
		self.EntryFrame.grid(row=2,column=0,pady=1,columnspan=2)


		Label(self.HardListFrame, text='Pick Hard parts:').grid(row=0,column=1)
		#self.HardScroll  = Scrollbar(self.HardListFrame, orient=VERTICAL)
		#self.HardScroll.grid( row=1, column=0,sticky=N+S+E)
		#self.HardBox  = Listbox(self.HardListFrame, height=15, selectmode=MULTIPLE,yscrollcommand = self.HardScroll.set,exportselection=0)
		self.HardBox  = Listbox(self.HardListFrame, height=15, width=36,selectmode=MULTIPLE,exportselection=0)
		self.HardBox.grid( row=1, column=1)
		#self.HardScroll["command"] = self.HardBox.yview

		Label(self.SoftListFrame, text='Pick one Soft part:').grid(row=0,column=1)
		#self.SoftScroll  = Scrollbar(self.SoftListFrame, orient=VERTICAL)
		#self.SoftBox  = Listbox(self.SoftListFrame, height=15, selectmode=SINGLE,yscrollcommand = self.SoftScroll.set,exportselection=0)
		self.SoftBox  = Listbox(self.SoftListFrame, height=15, width=36, selectmode=SINGLE,exportselection=0)
		self.SoftBox.grid( row=1, column=1)
		#self.SoftScroll["command"] = self.SoftBox.yview

		self.SoftScrollUp  = Button(self.SoftListFrame,text="^",command=self.ScrollSoftUpward)
		self.SoftScrollDn  = Button(self.SoftListFrame,text="v",command=self.ScrollSoftDnward)
		self.HardScrollUp  = Button(self.HardListFrame,text="^",command=self.ScrollHardUpward)
		self.HardScrollDn  = Button(self.HardListFrame,text="v",command=self.ScrollHardDnward)
		self.SoftScrollUp.grid( row=2, column=0)
		self.SoftScrollDn.grid( row=2, column=1)
		self.HardScrollUp.grid( row=2, column=0)
		self.HardScrollDn.grid( row=2, column=1)

		Label(self.EntryFrame, text='Smooth').grid(row=3,column=0)
		self.SmoothSlider = Scale(self.EntryFrame, from_=(-1), to=5,orient=HORIZONTAL,resolution=1,tickinterval=1)
		self.SmoothSlider.grid(row=3, column=1,columnspan=3)
		self.SmoothSlider.set(0)

		# buttons for Go, cancel.
		self.ButtonGo = Button(master, text='Go', command=self.HandleGo)
		self.ButtonGo.grid(row=6, sticky=W)
		self.ButtonCancel = Button(master, text='Cancel', command=self.HandleCancel)
		self.ButtonCancel.grid(row=6, sticky=E)

		self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

		# Fill the lists with available objects.
		# Figs first, as in Posette : Head, then plain actors.
		if scene.Figures():
			for OneFig in scene.Figures():
				FigName=OneFig.Name()
				for OneActor in OneFig.Actors():
					ActorName=OneActor.Name()
					if string.upper(ActorName) != 'BODY':
						st="%s:%s" % (FigName,ActorName)
						self.HardBox.insert(END,st)
						self.SoftBox.insert(END,st)
		for OneActor in scene.Actors():
			if OneActor.IsProp():
				ActorName=OneActor.Name()
				st="[prop]:%s" % ActorName
				self.HardBox.insert(END,st)
				self.SoftBox.insert(END,st)

	# - - - - - - - - - - - - - - - - - -
	def ScrollSoftUpward(self):
		self.SoftBox.yview_scroll(-1,'units')
	def ScrollSoftDnward(self):
		self.SoftBox.yview_scroll( 1,'units')
	def ScrollHardUpward(self):
		self.HardBox.yview_scroll(-1,'units')
	def ScrollHardDnward(self):
		self.HardBox.yview_scroll( 1,'units')
	# - - - - - - - - - - - - - - - - - -
	def ShowStatus(self,S): # Just saving repetition
		self.StatusEntry.delete(0,END)
		self.StatusEntry.insert(0,S)
		self.StatusEntry.update_idletasks()
	# - - - - - - - - - - - - - - - - - -

	def HandleGo(self):
		global SoftName
		global HardNames
		global SmoothOn
		# Fetch the Hard and Soft list selections.
		Sels=self.HardBox.curselection()
		if not Sels:
			self.ShowStatus("Must select some Hard Parts.")
			return
		else:
			for OneSel in Sels:
				ListString=self.HardBox.get(int(OneSel))
				# Parse the list string
				Parts=string.split(ListString,':')
				if not Parts: continue
				HardNames.append( (Parts[0],Parts[1]) )

		Sels=self.SoftBox.curselection()
		if not Sels:
			self.ShowStatus("Must select one Soft Part.")
			return
		else:
			ListString=self.HardBox.get(int(Sels[0]))
			# Parse the list string
			Parts=string.split(ListString,':')
			if not Parts:
				self.ShowStatus("Must select one Soft Part.")
			else: SoftName = ( (Parts[0],Parts[1]) )

		self.SmoothVar = self.SmoothSlider.get()

		self.ShowStatus("Working....")
		SetupVars()
		Changed = Displace()
		if self.SmoothVar:
			self.ShowStatus("Smoothing....")
			Smooth(self.SmoothVar)
		if Changed: Preserve()	# Always preserve
		scene.DrawAll()

		self.master.destroy()

	def HandleCancel(self):
		self.master.destroy()

#------------------------------------------------------------
# Activate the TK loop
#------------------------------------------------------------

app = App(root, '')
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
